/*
 *  Jomar3D, a lobster-running-to-catch game.
 *  Copyright (C) 2006-2008 Jonathan Michalon <studios.chalmion@no-log.org>.
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "sound.h"

#if (USESOUND)

#include <time.h>
#include "draw.h"
#include "jomar3D.h"
#include "utils.h"

#include <SDL.h>
#include <SDL_mixer.h>

#define MAXMUSIC 100



static Mix_Chunk* sound_loadFile    (const char *name);
static void       sound_init        (void);
static int        sound_switchMusic (void *dummy);


static int l_n_played = 0;            /* number of musics already played */
static int l_start_cycle_time = 0;    /* time at which a cycle of music was started */
static int l_files_done[MAXMUSIC];    /* flag to know a file was already played */
  
static int        l_n_music = 0;           /* number of available musics */
static char*      l_sound_files[MAXMUSIC]; /* name of each available music */
static Mix_Chunk* l_sounds[SOUND_N];       /* array of all loaded sounds */

static float      l_volume_set = 1.;       /* percent of the volume to use */


/* load all musics */
static int
sound_loadAll (void *dummy)
{
  (void) dummy;

  #define LOAD(snd, sndname) \
    (l_sounds[snd] = sound_loadFile (SOUNDDIR"/"sndname".wav"))

/* load the game's sounds */
  LOAD (SOUND_WALL,     "wall");
  LOAD (SOUND_TRYCATCH, "trycatch");
  LOAD (SOUND_PICKUP,   "pickup");
  LOAD (SOUND_WON,      "won");
  LOAD (SOUND_WALK,     "walk");
  LOAD (SOUND_MENU,     "menu");
  LOAD (SOUND_CHOU,     "chou");
  LOAD (SOUND_BOOM,     "boom");
  
  #undef LOAD
  
/* select, load and launch the first music */
  sound_switchMusic (NULL);
  
  return 1;
}

/* free all loaded musics */
void
sound_unloadAll (void)
{
  int i;
  
  for (i = 0; i < SOUND_N; i++)
  {
    if (l_sounds[i])
      Mix_FreeChunk (l_sounds[i]);
  }
  
  for (i = 0; i < MAXMUSIC; i++)
    free (l_sound_files[i]);
}

/* change the background music */
void
sound_changeMusic (void)
{
  if (core_checkState(JOMAR3D_PLAYMUSIC))
    Mix_FadeOutChannel (0, 1000);
}

/* launch all the music process */
void
sound_launch (void)
{
  int i;

  
  for (i=0; i < SOUND_N; i++)
    l_sounds[i] = NULL;
  
  if (core_checkState (JOMAR3D_CANMUSIC))
  {
    debugmsg (("Launching the sounds\n"));
    /* init the sound device */
    sound_init ();
    
    /* define the initial volume setting */
    core_initVolume ();
    
    /* start all */
    if (core_checkState (JOMAR3D_CANMUSIC)) /* could have changed */
      SDL_CreateThread (sound_loadAll, NULL);
  }
}

/* mix a sound given its index number */
void
sound_mix (int soundnb)
{
  if (!core_checkState(JOMAR3D_PLAYMUSIC))
    return;
  
  debugmsg (("mixing sound number %d\n", soundnb));
  
# define MIX(name) \
  do { \
    if (l_sounds[name]) \
      if (Mix_Playing (soundnb) == 0) \
        Mix_PlayChannel (soundnb, l_sounds[name], 0); \
  } while (0)

  switch (soundnb)
  {
    case WALL:
      MIX (SOUND_WALL);
      break;
    case TRYCATCH:
      MIX (SOUND_TRYCATCH);
      break;
    case PICKUP:
      MIX (SOUND_PICKUP);
      break;
    case WON:
      MIX (SOUND_WON);
      break;
    case WALK:
    case BOOST:
      MIX (SOUND_WALK);
      break;
    case CHOU:
      MIX (SOUND_CHOU);
      break;
    case BOOM:
      MIX (SOUND_BOOM);
      break;
    case SMENU1:
    case SMENU2:
    case SMENU3:
    case SMENU4:
    case SMENU5:
    case SMENU6:
    case SMENU7:
    case SMENU8:
    case SMENU9:
    case SMENU10:
      MIX (SOUND_MENU);
      break;
    
    default:
      printf ("No sound %d to mix\n", soundnb);
  }
#undef MIX
}

/* Stop a particuliar channel of the mixer */
void
sound_stopChannel (int channel)
{
  if (core_checkState (JOMAR3D_CANMUSIC))
    Mix_HaltChannel (channel);
}

/* Invert the current state of the sound: pause or resume */
void
sound_pauseResume (void)
{
  if (core_checkState (JOMAR3D_CANMUSIC))
  {
    core_invertState (JOMAR3D_PLAYMUSIC);
    if (core_checkState(JOMAR3D_PLAYMUSIC))
      Mix_Resume (-1);
    else
      Mix_Pause (-1);
  }
}

/* Set the volume of all channel */
static void
sound_setVolume (void)
{
  int n;

  /* Prevent to call the SDL_mixer when no sound is enabled */
  if (!core_checkState (JOMAR3D_CANMUSIC))
    return;
  /* change all channel's volume */
  Mix_Volume (-1, MIX_MAX_VOLUME * l_volume_set);
  /* re-set the menu's correctly */
  for (n = SMENU1; n <= SMENU10; n++)
    Mix_Volume (n, MIX_MAX_VOLUME * (1 - (n - SMENU1) * 0.15) * l_volume_set);
}

/* Change the volume of the music */
void
sound_changeVolume (int flag, float percent)
{
  switch (flag)
  {
    case S_SET:
      l_volume_set = percent;
      break;
    case S_DECREASE:
      l_volume_set -= 0.1;
      if (l_volume_set < 0.)
        l_volume_set = 0.;
      break;
    case S_INCREASE:
      l_volume_set += 0.1;
      if (l_volume_set > 1.)
        l_volume_set = 1.;
      break;
  }
  sound_setVolume ();

/* write the changed value (only if changed and not just set by init) */
  if (flag == S_DECREASE || flag == S_INCREASE)
    core_saveOptions ();

  debugmsg (("Changed volume to %.1f%%\n", l_volume_set));
}

/* return the current volume */
float
sound_getVolume (void)
{
  return l_volume_set;
}

/* load a music file */
static Mix_Chunk*
sound_loadFile (const char *name)
{
  Mix_Chunk* ptr;
  
  debugmsg (("loading %s\n", name));
  
  /* load the song (wav, ogg, voc) */
  if ((ptr = Mix_LoadWAV (name)) == NULL)
  {
    fprintf (stderr, "error in Mix_LoadWAV(%s): %s\n", name, SDL_GetError ());
  }
  
  return ptr;
}

/* switch between the musics */
static int
sound_switchMusic (void *dummy)
{
  int i;
  int playnb;                 /* number of the music which is going to be played */
  char *ptr;
  (void) dummy;

  if (l_start_cycle_time == 0)
    l_start_cycle_time = time (NULL);
  
  debugmsg (("startcycletime %d\n", l_start_cycle_time));
  debugmsg (("played %d\n", l_n_played));
  
  if (l_n_music == 0)
    return 1;
  
another:;
  if (l_n_played >= l_n_music)
  {
    debugmsg (("new cycle \n"));
    
    for (i = 0; i < l_n_music; i++)
      l_files_done[i] = FALSE;
    
    l_n_played = 0;
    if (time (NULL) - l_start_cycle_time <= 5)
    {
      fprintf (stderr, "Too short time in one cycle of the music\n");
      return 1;
    }
    l_start_cycle_time = time (NULL);
  }
  playnb = MYRAND * l_n_music;
  
  if (l_files_done[playnb])
    goto another;
  
  debugmsg (("playnb %d\n", playnb));
  
  l_files_done[playnb] = TRUE;
  l_n_played++;
  ptr = malloc ((strlen (MUSICDIR "/") + strlen (l_sound_files[playnb]) + 1));
  if (ptr)
  {
    sprintf (ptr, MUSICDIR "/%s", l_sound_files[playnb]);
    /* if a previous music was loaded, free it before loading the next one */
    if (l_sounds[SOUND_BACKGROUND])
    {
      Mix_FreeChunk (l_sounds[SOUND_BACKGROUND]);
    }
    /* try to load a music and retry with another on failure */
    if (!(l_sounds[SOUND_BACKGROUND] = sound_loadFile (ptr)))
    {
      free (ptr);
      goto another;
    }
  }
  else
    perror ("failed to allocate memory");
  free (ptr);
  
  debugmsg (("playing %s\n", l_sound_files[playnb]));

  draw_setMusicName (l_sound_files[playnb]);

  /* Let's launch a background music */
  if (Mix_PlayChannel (0, l_sounds[SOUND_BACKGROUND], 0) == -1)
    fprintf (stderr, "Mix_PlayChannel: %s\n", Mix_GetError ());
  /* If we want no background music, pause it now that it is registered */
  if (!core_checkState (JOMAR3D_PLAYMUSIC))
    Mix_Pause (0);

  /*if (ratio != 0.)*/
    draw_forceRedraw ();

  return 1;
}

/* SDL_Mixer called routine when a channel has finished playing */
static void
sound_onChannelEnded (int channel)
{
  debugmsg (("Channel %d finished\n", channel));
  
  if (channel == 0)
    SDL_CreateThread (sound_switchMusic, NULL);
}

/* filter to select the right file's extensions */
static int
sound_musicFilter (const struct dirent *a)
{
  const char *ext;
  
  debugmsg (("checking: %s\n", a->d_name));
  
  ext = getext (a->d_name);
  
#define CHECK(str) \
  if (streq (ext, str)) return 1
  
  if (ext)
  {
    CHECK ("ogg");
    CHECK ("wav");
    CHECK ("voc");
  }
  
#undef CHECK
  
  return 0;
}

/* initialize the sound playing process */
static void
sound_init (void)
{
  struct dirent **namelist;
  int n;
  
  /* Load the SDL library */
  if (SDL_Init (SDL_INIT_AUDIO) < 0)
  {
    fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
    core_deleteState (JOMAR3D_CANMUSIC);
    core_deleteState (JOMAR3D_PLAYMUSIC);
    return;
  }
  atexit (SDL_Quit);
  
  /* Open the sdl_mixer audio device so loading on-the-fly is possible */
  if (Mix_OpenAudio (44100, MIX_DEFAULT_FORMAT, 2, 2048) < 0)
  {
    fprintf (stderr, "Couldn't open audio: %s\n", Mix_GetError ());
    core_deleteState (JOMAR3D_CANMUSIC);
    core_deleteState (JOMAR3D_PLAYMUSIC);
    SDL_Quit ();
    return;
  }
  atexit (Mix_CloseAudio);
  
  Mix_AllocateChannels (LAST);
  sound_setVolume ();
  Mix_ChannelFinished (sound_onChannelEnded);
  
  l_n_music = myscandir (MUSICDIR, &namelist, sound_musicFilter);
  if (l_n_music == -1)
  {
    perror ("scandir");
    printf ("Error listing the music directory "MUSICDIR
            "'. You will probably experience problems...\n");
    l_n_music = 0;
    return;
  }
  if (l_n_music > MAXMUSIC)
    l_n_music = MAXMUSIC;

  /* Init the filename pointers */  
  for (n = 0; n < MAXMUSIC; n++)
    l_sound_files[n] = NULL;

  for (n = 0; n < l_n_music; n++)
    {
      l_sound_files[n] = malloc (strlen (namelist[n]->d_name) + 1);
      if (l_sound_files[n])
        sprintf (l_sound_files[n], "%s", namelist[n]->d_name);
      else
        perror ("failed to allocate memory");
      
      debugmsg (("%s\n", l_sound_files[n]));
    }
  
  n = l_n_music;
  while (n--)
    free (namelist[n]);
  free (namelist);
  
  debugmsg (("number of musics: %d\n", l_n_music));
}
#endif /* (USESOUND) */
